# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Plugin module which creates a rotating set of gzipped csv output files """

import os

import logging
import time
import traceback

import subprocess
import threading

import datetime
from datetime import timezone
from dateutil import parser
from dateutil.tz import UTC

import glob
import shutil
import re
import math
import pickle
import csv
import itertools

import filter_ingest

from foglamp.common import logger
from foglamp.common.plugin_helpers import HandleMap, PluginHandle

# local logger
_LOGGER = logger.setup(__name__, level=logging.INFO)

__author__ = "Douglas Orr and Deepanshu Yadav"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

FOGLAMP_DATA_ENV = "FOGLAMP_DATA"
FOGLAMP_DATA_PFX = "FOGLAMP_DATA/"
FOGLAMP_DEFAULT_DATA_DIR = "/usr/local/foglamp/data"


class ManagementThread(threading.Thread):
    """ ManagementThread - compress / remove extra files as a background process """

    def __init__(self, parent, suffix):
        super().__init__(daemon=True)  # establish thread
        self.logger = logger.setup("{}-manager".format(suffix), level=logging.INFO)

        self.parent = parent
        self.suffix = suffix

        self.logger.info("MANAGEMENT thread started")

    def run(self):
        p = self.parent

        while True:
            with p.work_cond:
                self.logger.debug("MANAGEMENT thread: sleep cron mode {}".format(p.cron_mode))

                # check for new work
                to_compress = p.to_compress
                p.to_compress = []
                if len(to_compress) == 0:
                    # release the lock and do the queued work
                    if p.shutting_down:
                        # no more work and we're trying to exit
                        self.logger.info("SHUTTING DOWN MANAGER for cron mode {}".format(p.cron_mode))
                        break

                    p.work_cond.wait()
                    self.logger.debug("MANAGEMENT thread: wake cron mode {}".format(p.cron_mode))

            if len(to_compress) != 0:
                self.logger.debug("MANAGEMENT thread: found {} to compress no of files are {}".
                                  format(to_compress, len(to_compress)))
                # compress could take a while, don't hold the lock
                self._compress(to_compress)
                with p.work_cond:
                    if p.max_directories is not None:
                        self._rotate()

    def _compress(self, paths):
        p = self.parent
        compression_type = p.compression_type
        encryption_enabled = False

        self.logger.debug(" compress type: {}".format(compression_type))

        if compression_type == 'bzip2':
            # gzip all available csv's; wait for bzip to finish; do newest
            cmd = ['bzip2', '-f'] + sorted(paths, reverse=True)

        elif compression_type == 'gzip':
            # gzip all available csv's; wait for bzip to finish; do newest
            cmd = ['gzip', '-f'] + sorted(paths, reverse=True)

        elif compression_type == '7za':
            encryption_enabled = True

            if p.encrypt_pw == '':
                raise ValueError("missing encryption password")
            for file_name in sorted(paths):
                target_file_name = os.path.splitext(file_name)[0]
                cmd = ['7z', 'a', '-y', '-p{}'.format(p.encrypt_pw), '-sdel', '-mx=6', '-mhe=on', '-t7z',
                       target_file_name, file_name]
                # a -> add, y->yes, -p->password, -sdel-> delete original file after operation,
                # mx->compression level(1-9) , mhe-> enable header encryption, -t -> compression format

                self.logger.debug("cmd is {}", cmd)
                try:
                    p = subprocess.run(cmd, stderr=subprocess.PIPE)
                    if p.returncode != 0:
                        _err = p.stderr.decode('utf-8').strip()
                        # if there was error output, log it
                        self.logger.error("compress: runtime error")
                        for line in _err.split('\n'):
                            self.logger.error(line)
                except Exception as e:
                    self.logger.exception("compress failure with exception {}".format(e))
        else:
            raise ValueError("{}: unknown compression type".format(compression_type))

        if not encryption_enabled:
            self.logger.debug("COMPRESSING {}".format(cmd))
            try:
                p = subprocess.run(cmd, stderr=subprocess.PIPE)
                if p.returncode != 0:
                    _err = p.stderr.decode('utf-8').strip()
                    # if there was error output, log it
                    self.logger.error("compress: runtime error")
                    for line in _err.split('\n'):
                        self.logger.error(line)
            except Exception as e:
                self.logger.exception("compress failure with exception {}".format(e))

    def _rotate(self):
        """ _rotate - if we have more than max_directories (compressed or not) remove the extra """
        p = self.parent
        _dir_list = glob.glob("{}/*".format(p.dest_dir))
        _expr = lambda x: datetime.datetime.strptime(os.path.split(x)[1].split(".")[0], "%Y-%m-%d")
        to_rotate = sorted(_dir_list, key=_expr)
        if p.max_directories > 0:
            while len(to_rotate) > p.max_directories:
                try:
                    # if we have the max files possible, rotate
                    _rm_dir = to_rotate.pop(0)
                    self.logger.debug("MANAGEMENT - ROTATE: removing {}".format(_rm_dir))
                    shutil.rmtree(_rm_dir)
                except:
                    traceback.exc_print()
        else:
            self.logger.info("Max directories to keep are zero. Not deleting anything.")


class MediaHandler(PluginHandle):
    """
    Stores state passed to/from FogLAMP callbacks to manage periodic forwarding of data
    """

    def __init__(self, config, suffix, ingest_ref, callback):
        super().__init__("{}-output".format(suffix), (ingest_ref, callback))

        self.logger.setLevel(logging.INFO)
        self.logger.info("media-handler started")

        # slots
        self.forward_data = None
        self.input_assets = None
        self.dest_dir = None
        self.filter_name = None
        self.add_timestamp = None
        self.cron_mode = None
        self.cron_period_start = None
        self.cron_tab_spec = None
        self.cron_period_duration = None
        self.event_pre_time = None
        self.event_repetition_time = None
        self.event_duration = None
        self.event_post_time = None
        self.rotate_after = None
        self.clean_data = None
        self.dynamic_sampling_rate = None
        self.sampling_rate = None
        self.max_directories = None
        self.file_type = None
        self.enable_compress = None
        self.compression_type = None
        self.encrypt_pw = None

        self.config_update(config)  # transfer config entries to handle slots

        self.assets = None
        if self.input_assets != '':
            self.assets = self.input_assets.strip().split(",")

        self.suffix = suffix
        self._now = None

        if self.dest_dir.startswith(FOGLAMP_DATA_PFX):
            _foglamp_data_dir = os.getenv(FOGLAMP_DATA_ENV, default=FOGLAMP_DEFAULT_DATA_DIR)
            self.dest_dir = os.path.join(_foglamp_data_dir, self.dest_dir[len(FOGLAMP_DATA_PFX):], self.filter_name)
        else:
            self.dest_dir = os.path.join(self.dest_dir, self.filter_name)

        if not os.path.exists(self.dest_dir):
            os.makedirs(self.dest_dir)

        # ### state ###
        self.cron_start = None
        self.n_records = self.max_records = 0  # to be established in parse_cron()

        self.is_cron_start_initialized = False
        self.is_cron_parsed = False
        self.period_delta = None
        self.event_repetition_delta = None
        self.start_collection = False
        self.has_continuous_collection_started = False

        self.ext = 0
        self.cur_file_name = None
        self.media_file = None

        # ### management thread ### #
        self.work_cond = threading.Condition()
        self.to_compress = []  # queue of files that need to be compressed
        self.shutting_down = False

        self.file_work_thread = ManagementThread(self, suffix)
        self.file_work_thread.start()

        self._handle_unprocessed_files()

    # -------------------PUBLIC API-----------------------
    def ingest(self, data):
        """ ingest - save to medium if enabled, pass along to subsequent filters/north as appropriate """

        if not self.enable or self.forward_data:
            # pass data upstream if not filtering, or if explicit forward_data flagged
            filter_ingest.filter_ingest_callback(self.callback, self.ingest_ref, data)

        if self.enable:
            if type(data) is dict:
                data = [data]

            if self.assets:
                # only save our chosen assets
                data = list(filter(lambda d: (d['asset'] in self.assets), data))

            if data != []:
                # make sure we update the now variable every time the new batch arrives
                # even if we are not collecting readings

                if self.fields is None:
                    # very first time through, make a note of our fields. ASSUME this doesn't change
                    self.fields = list(data[0]['readings'].keys())
                    self.ts_col = 'user_ts' if 'user_ts' in data[0] else (
                        'timestamp' if 'timestamp' in data[0] else None)

                    if not self.is_cron_parsed:
                        self._parse_cron()
                        # first time through. get a target cron start and new media file
                        self.logger.debug("The cron start is {}".format(self.cron_start))
                        self.is_cron_parsed = True

                    if not self.is_cron_start_initialized and self.cron_period_start == "":
                        self.cron_start = parser.parse(data[0][self.ts_col])
                        self.logger.info(
                            "The cron start is {} and the now variable is {} and cron mode is {}".format(
                                self.cron_start, self._now, self.cron_mode))
                        if self.cron_mode == "continuous":
                            # start the collection immediately for continuous
                            # as well as create a file for writing the data.
                            self.start_collection = True
                            self.has_continuous_collection_started = True
                            self._new_media_file()
                        else:
                            # updating the cron start to the next scheduled time
                            # if we are in periodic or table cron mode.
                            self._update_cron_time()
                    self.is_cron_start_initialized = True

                if self.ts_col:

                    # updating the now variable with the
                    # latest timestamp from reading rather than the latest system time.
                    # only for table mode
                    if self.cron_mode == "table":
                        self._now = parser.parse(data[-1][self.ts_col])

                    _previous_status_of_collection = self.start_collection
                    self.start_collection, index = self.should_start_collection(data)

                    if not _previous_status_of_collection and self.start_collection:
                        # time to create new file
                        # so creating new file only when it is actually required.
                        self.logger.info("Switching on collection and creating new file.")
                        self._new_media_file()

                    if self.start_collection:
                        if self.cron_mode == "continuous":
                            self.has_continuous_collection_started = True
                        self.save(self.extract_rows(data[index:]))
                else:
                    self.logger.error("The time stamp field is not present in the data")

    def save(self, data_iter):
        """ save - save data generated by data_iter, if max_records is reached, rotate files """

        if data_iter == []:
            # filter is finished. force close
            if self.media_file:
                del self.media_file
                self.media_file = None
                self.n_records = 0
                return

        cur_file_name_ts_str = os.path.split(self.cur_file_name)[-1]
        cur_file_name_ts_str2 = "-".join(cur_file_name_ts_str.split('-')[:-1])
        self.logger.debug("cur_file_name_ts_str={}, cur_file_name_ts_str2={}".format(cur_file_name_ts_str, cur_file_name_ts_str2))
        cur_file_name_ts = datetime.datetime.strptime(cur_file_name_ts_str2, "%Y-%m-%d-%H-%M-%S").replace(tzinfo=timezone.utc)  # .astimezone("India/Kolkata")

        _list = []

        # write out the generated data in chunks
        for i,d in enumerate(data_iter):
            if self.cron_mode == "continuous" and self.dynamic_sampling_rate:
                if i==0 and not isinstance(d[-1], str):
                    self.logger.error("When cron_mode is 'continuous' and dynamic_sampling_rate=True, timestamp must be present in readings, else enable 'Add timestamp to csv data' in csv writer configuration'")

                self.logger.debug("Checking whether current file should be closed: d={}".format(d))
                curr_record_ts = parser.parse(d[-1]).astimezone(UTC)
                self.logger.debug("curr_record_ts={}, cur_file_name_ts={}".format(curr_record_ts, cur_file_name_ts))

                end_file = ((curr_record_ts - cur_file_name_ts) >= self.period_delta)
                if end_file:
                    self.logger.info("Checking whether current file should be closed: d={}, curr_record_ts={}, cur_file_name_ts={}, self.period_delta={}, end_file={}, self.ts_col={}, self.cur_file_name={}, self.n_records={}, self.max_records={}"
                            .format(d, curr_record_ts, cur_file_name_ts, self.period_delta, end_file, self.ts_col, self.cur_file_name, self.n_records, self.max_records))
            else:
                end_file = (self.n_records == self.max_records)
                
            if end_file:
                # file is "full" , flush and start a new file.
                if len(_list) > 0:
                    self.write_rows(_list);
                    _list = []
                self.logger.info("Closing the current file {}".format(self.cur_file_name))

                # when maximum records are reached.
                # 1. Close the current file.
                # 2. Update the cron time.
                # 3. If continuous mode then create new file
                # 4. If periodic or table mode then only switch off collection and wait
                # for next scheduled time.

                self._close_cur_file()
                self._update_cron_time()
                if self.cron_mode == "continuous":
                    self.logger.info("creating new file for continuous mode")
                    self._new_media_file()  # create new file based on cron_start
                    cur_file_name_ts_str = os.path.split(self.cur_file_name)[-1]
                    cur_file_name_ts_str2 = "-".join(cur_file_name_ts_str.split('-')[:-1])
                    self.logger.info("cur_file_name_ts_str={}, cur_file_name_ts_str2={}".format(cur_file_name_ts_str, cur_file_name_ts_str2))
                    cur_file_name_ts = datetime.datetime.strptime(cur_file_name_ts_str2, "%Y-%m-%d-%H-%M-%S").replace(tzinfo=timezone.utc)

                if self.cron_mode != "continuous":
                    self.logger.info("Switching off collection for {} mode ".format(self.cron_mode))
                    self.start_collection = False

            if self.start_collection:
                _list.append(d)
                self.n_records += 1

        if len(_list) > 0:
            # generator has paused, flush anything that built up
            self.write_rows(_list)

    def shutdown(self):
        self.logger.info("Closing the current file.")
        self._close_cur_file()
        self.logger.info(" The current file has been closed.")

        with self.work_cond:
            self.logger.info("Stopping the management thread.")
            self.shutting_down = True
            self.logger.info("Notifying the worker thread .")
            self.work_cond.notify()

        self.logger.info("Joining the management thread.")
        self.file_work_thread.join()
        self.logger.info("Joined the management thread.")

    def now(self) -> datetime.datetime:
        _now = self._now if self._now else datetime.datetime.now(timezone.utc)
        return _now

    def cron_next(self) -> datetime.datetime:
        """ cron_next - find the next time we'll start recording, based on our cron mode and spec """

        if self.cron_mode == "table":
            return self._cron_table_next()
        elif self.cron_mode == "periodic":
            # our next start time is now, plus the next cycle
            return self.cron_start + self.event_repetition_delta
        elif self.cron_mode == "continuous":
            return self.cron_start + self.period_delta

        return self.cron_start

    def ceilSearch(self, arr, low, high, x):

        # In this function we intend to find the smallest timestamp index 
        # that is greater than or equal to the given timestamp (x).
        # we are asumming that the timestamps never repeat.

        # sanity check
        if low > high:
            return -1

        # This case won't happen already made sure earlier. 
        if x > parser.parse(arr[high][self.ts_col]):
            return -1
        
        # the first element is the required element in this case as x is less than 
        # first element.
        if x <= parser.parse(arr[low][self.ts_col]):
            return low

        mid = (low + high) // 2

        # return the mid element
        if x == parser.parse(arr[mid][self.ts_col]):
            return mid
        
        # if true then x lies somewhere over right of mid
        elif x > parser.parse(arr[mid][self.ts_col]):

            # just make sure that element just greater than mid 
            # is not the required element. if it is then return it 
            # otherwise search in the right.
            if mid+1 <= high and x <= parser.parse(arr[mid+1][self.ts_col]):
                return mid + 1
            else:
                return self.ceilSearch(arr, mid+1, high, x)
        else:
            # if true then the x lies somewhere to left of mid

            # we will return mid if the element before mid is less than x
            if mid-1 >= low and x > parser.parse(arr[mid-1][self.ts_col]):
                return mid
            # otherwise we serach in the left.
            else:
                return self.ceilSearch(arr, low, mid-1, x)
        
    def search_for_index(self, data):

        # we will call the routine ceilSearch for given data. 
        # The routine will return an index of the data array from which
        # collection should begin.  
        return self.ceilSearch(data, 0, len(data) - 1, self.cron_start)

    def should_start_collection(self, data):
        """
            Routine that checks whether to start collection or not.
            If yes then at which reading number we should collecting

            Args
                data : The readings array

            Returns : 1.True if the collection should start False if collection should not begin.
                      2. The index (inside the batch of readings) from where collection should begin.
        """

        # if mode is continuous and collection has begun atleast once
        # then no need to check and simply return True. Otherwise we may need to check.
        if self.cron_mode == "continuous" and self.has_continuous_collection_started:
            return True, 0

        # once the collection has begun keep returning True.
        # We will check only start_collection becomes False.
        # And that can only happen if collection limit for a file
        # reaches the desired value.
        if not self.start_collection:

            # the required time is the time of last reading in this batch.
            last_row_time = parser.parse(data[len(data) - 1][self.ts_col])
            last_row_time.astimezone(UTC)

            # if this time of last reading is less than cron time then do not begin collection.
            # will wait for other batch to come.
            if last_row_time < self.cron_start:
                return False, -1
            # The last reading is equal to cron time begin the collection and include the last reading.
            elif last_row_time == self.cron_start:
                return True, len(data) - 1
            else:
                return True, self.search_for_index(data)
        else:
            # collection already started , wait till it becomes False.
            return True, 0

    # # -------------------PRIVATE API-----------------------

    def _handle_unprocessed_files(self):
        """
            Checks that if any uncompressed csv files are not there. Deletes them if empty
            and compresses them if contains something.
        Returns:
            None
        """
        _compression_file_list = []
        for root, dirs, files in os.walk(self.dest_dir):
            for f in files:
                _full_file_path = os.path.join(root, f)
                if _full_file_path.endswith(('.csv')):
                    if os.stat(_full_file_path).st_size == 0:
                        self.logger.info("Removing empty file {}".format(_full_file_path))
                        os.remove(_full_file_path)
                    else:
                        self.logger.info("sending raw file {} for compression.".format(_full_file_path))
                        _compression_file_list.append(_full_file_path)

        if _compression_file_list:
            with self.work_cond:
                self.to_compress = _compression_file_list
                self.work_cond.notify()

    def _parse_cron(self):
        # parse cron values (whose validity depends on cron_mode)

        # duration of the "period" during which we record to a file
        if self.cron_mode == 'continuous' or self.cron_mode == 'table':
            _period_sec, _period_microsecond = self._parse_to_ms(self.cron_period_duration)
            self.period_delta = datetime.timedelta(seconds=_period_sec, microseconds=_period_microsecond)
            self.max_records = _period_sec * self.sampling_rate + \
                               ((_period_microsecond * self.sampling_rate) / 1000000)

        # when "periodic" mode, we record for "period", then skip to the next cycle
        if self.cron_mode == 'periodic':

            _pre_sec, _pre_microsecond = self._parse_to_ms(self.event_pre_time)
            self.pre_delta = datetime.timedelta(seconds=_pre_sec, microseconds=_pre_microsecond)

            _event_duration_sec, _event_duration_microsecond = self._parse_to_ms(self.event_duration)
            self.event_duration_delta = datetime.timedelta(seconds=_event_duration_sec,
                                                           microseconds=_event_duration_microsecond)
            _post_sec, _post_microsecond = self._parse_to_ms(self.event_post_time)
            self.post_delta = datetime.timedelta(seconds=_post_sec, microseconds=_post_microsecond)

            self.total_collection_delta = self.pre_delta + self.event_duration_delta + self.post_delta

            _event_repetition_sec, _event_repetition_microsecond = self._parse_to_ms(self.event_repetition_time)
            self.event_repetition_delta = datetime.timedelta(seconds=_event_repetition_sec,
                                                             microseconds=_event_repetition_microsecond)

            if self.event_repetition_delta < self.total_collection_delta:
                raise ValueError("The periodic duration {} should always be greater than collection duration{}".format(
                    self.event_repetition_delta, self.total_collection_delta))

            self.max_records = self.total_collection_delta.seconds * self.sampling_rate + \
                               (self.total_collection_delta.microseconds * (self.sampling_rate / 1000000))

        # total duration of the session
        _total_sec, _total_microsecond = self._parse_to_ms(self.rotate_after)

        # Setting the max_directories for collection
        total_seconds = round(_total_sec + (_total_microsecond / 1000000))
        self.max_directories = math.ceil(total_seconds / 86400)  # 1 day is 86400 seconds.
        # Calculating how many days worth of data going to store.

        # we need to preserve the data even after the designated time. So increasing by 1.
        self.max_directories += 1

        if self.cron_period_start != "":
            _start = parser.parse(self.cron_period_start)
            _start = _start.astimezone(UTC)

            # subtract the pre collection time
            if self.cron_mode == "periodic":
                _start = _start - self.pre_delta

            self.cron_start = _start
            self._now = datetime.datetime.now(tz=timezone.utc)
            if _start < self._now and self.cron_mode == "periodic":
                # the time given belongs to past
                # we wish to find out at what time this time past time
                # should be extended.
                _total_cycle_second = int(self.event_repetition_delta.seconds +
                                          (self.event_repetition_delta.microseconds / 1000000))
                _remaining_time = int((self._now - self.cron_start).total_seconds())
                _cycles = math.ceil(_remaining_time / _total_cycle_second)
                _actual_start_seconds = _cycles * _total_cycle_second

                _actual_start_delta = datetime.timedelta(seconds=_actual_start_seconds)
                self.cron_start = self.cron_start + _actual_start_delta

            self.logger.debug("The cron start is {} the now variable is {} and the "
                              "cron mode is {} ".format(self.cron_start, self._now, self.cron_mode))

        if self.cron_mode == 'table':
            # crontable has logic all its own
            if self.cron_tab_spec == '':
                raise ValueError("cronTable mode requires a cronSpec")
            self._init_cron_spec()  # may override max_records

        if self.cron_mode == "periodic" and self.event_repetition_delta == self.total_collection_delta:
            self.logger.debug("Collection time delta {} is equal to repetition time delta {}."
                              "So it is actually continuous collection. "
                              "Will treat it like continuous.".format(self.total_collection_delta,
                                                                      self.event_repetition_delta))
            self.cron_mode = "continuous"
            self.period_delta = self.total_collection_delta

        if self.cron_period_start != "":
            # creating a new file for continuous even if cron start is given.
            if self.cron_mode == "continuous":
                self._new_media_file()
            # for table cron mode we need to align the cron start with the cron string given so
            # we are updating the cron time.
            elif self.cron_mode == "table":
                self._update_cron_time()

        self.logger.info("the cron mode is {} the cron start is {}, max directories {} "
                         "max records {} period_delta {} , "
                         "event repetition delta {}".format(self.cron_mode, str(self.cron_start),
                                                            self.max_directories,
                                                            self.max_records,
                                                            self.period_delta,
                                                            self.event_repetition_delta))

    def _cron_table_next(self) -> datetime.datetime:
        """ cron_table_next - "table" mode, find the next starting time from our table """

        _hours = self.cron_dict.get('hour', None)
        _months = self.cron_dict.get('month', None)
        if 'day_of_week' in self.cron_dict:
            _weekdays = self.cron_dict.get('day_of_week', None)
            _days = None
        else:
            _days = self.cron_dict.get('day_of_month', None)
            _weekdays = None
        _minutes = self.cron_dict.get('minute', None)
        _seconds = self.cron_dict.get('second', None)

        _cur = self.now()

        # start looking at the next second after the current one
        _cur += datetime.timedelta(seconds=1)
        _cur = _cur.replace(microsecond=0)

        if _seconds:
            # find the next second candidate, maybe rounding to the next m/h/d/mo...
            while _cur.second not in _seconds:
                _cur += datetime.timedelta(seconds=1)

        if _minutes:
            _bumped = False
            while _cur.minute not in _minutes:
                _cur += datetime.timedelta(minutes=1)
                _bumped = True
            if _bumped:
                _cur = _cur.replace(second=0 if not _seconds else _seconds[0])

        if _hours:
            _bumped = False
            while _cur.hour not in _hours:
                _cur += datetime.timedelta(hours=1)
                _bumped = True
            if _bumped:
                _cur = _cur.replace(second=0 if not _seconds else _seconds[0],
                                    minute=0 if not _minutes else _minutes[0])

        if _days or _weekdays:
            _bumped = False
            if _days:
                while _cur.day not in _days:
                    _cur += datetime.timedelta(days=1)
                    _bumped = True
            elif _weekdays:
                while _cur.weekday() not in _weekdays:
                    _cur += datetime.timedelta(days=1)
                    _bumped = True
            if _bumped:
                _cur = _cur.replace(second=0 if not _seconds else _seconds[0],
                                    minute=0 if not _minutes else _minutes[0],
                                    hour=0 if not _hours else _hours[0])

        if _months:
            _bumped = False
            while _cur.month not in _months:
                _month = (_cur.month % 12) + 1
                _year = _cur.year + 1 if _month == 1 else _cur.year
                _cur = datetime.datetime(year=_year, month=_month, day=1, hour=0, minute=0, second=0)
                _bumped = True
            if _bumped:
                if _days:
                    _first_day = _days[0]
                elif _weekdays:
                    _first_day = _weekdays[0]
                else:
                    _first_day = 1
                _cur = _cur.replace(second=0 if not _seconds else _seconds[0],
                                    minute=0 if not _minutes else _minutes[0],
                                    hour=0 if not _hours else _hours[0],
                                    day=_first_day)

        self.logger.debug("cron next returns {}".format(_cur))
        return _cur

    def _init_cron_spec(self):
        """ init_cron_spec - one time initialization from cron table config specification """

        _cron = self.cron_tab_spec.strip().split(' ')
        if len(_cron) != 7:
            raise ValueError("Cron specifications must have seven space-separated components")
        self.cron_dict = {}

        _duration = float(_cron.pop())  # last entry is a simple float

        for k, v, r in zip(['second', 'minute', 'hour', 'day_of_month', 'month', 'day_of_week'],
                           _cron,
                           [(0, 59), (0, 59), (0, 23), (1, 31), (1, 12), (1, 7)]):
            if v == '*':
                continue
            elif '-' in v:
                # range <v1>-<v2>
                _r = list(map(int, v.split('-')))
                if len(_r) != 2: raise ValueError("cron ranges must be <int>-<int>")
                if _r[0] > _r[1]: raise ValueError("cron range start greater than range end")
                _v = list(range(_r[0], _r[1] + 1))
            elif ',' in v:
                # list of acceptable values
                _v = sorted(map(int, v.split(',')))
            else:
                # single value
                _v = [int(v), int(v)]

            if _v[0] < r[0] or _v[-1] > r[1]:
                raise ValueError("cron values {} outside of legal range {}".format(_v, r))

            self.cron_dict[k] = _v

        if 'day_of_month' in self.cron_dict and 'day_of_week' in self.cron_dict:
            raise ValueError("can't specify both day of month and day of week in cron spec")

        if _duration > 0:
            self.max_records = int(_duration * self.sampling_rate)

    def _new_media_file(self):
        """ new_media_file - previous file filled, open a file where the next record saved can go
            if cronSpec is set, open a file appropriate to the next time slot when records will be saved
            if cronSpec is not set, open a file marked by current time
        """

        with self.work_cond:
            _cur_fname, _cur_fd = self._cron_file()

            self.work_cond.notify()  # look for csvs to rotate

        self.logger.debug("NEW: created {}".format(_cur_fname))

        self.media_file = _cur_fd
        self.cur_file_name = _cur_fname
        self.n_records = 0

        # write the header row in the file immediately after creation
        _ts = ['user_ts'] if self.add_timestamp else []
        self.media_file.writerow(self.fields + _ts)

    # if cronSpec is set, max_records will be whatever is appropriate for the duration of
    # the calendar time that's been set; new_media_file will create a new subdirectory for
    # whenever the day changes and a maximum number of days' entries will be saved.

    # if cronSpec is not set, max_records is used as the signal for when a file is full,
    # but a full file just means that a new file is immediately started.

    def _cur_time_file(self):
        """ cur_time_file - LOCKED, create a new media file based on current time """
        self.logger.debug("NEW MEDIA TIME FILE")

        _hms = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        return self._media_file(_hms)

    def _update_cron_time(self):
        """
        Updates the cron start variable with the next scheduled time.
        Returns:
            None
        """
        self.logger.debug("Updating the cron time {}".format(self.cron_start))
        self.cron_start = self.cron_next()
        self.logger.info("Cron time has been updated to {}".format(self.cron_start))

    def _cron_file(self):
        """ cron_file - LOCKED, create a new media file for next time when samples are to be recorded """

        # dbo self.cron_start = self.cron_next()  # set up from next cron firing
        self.logger.info("FILE: set next time to be {}".format(self.cron_start))

        _hms = self.cron_start.strftime("%Y-%m-%d-%H-%M-%S")

        return self._media_file(_hms, day_dir="{}-{:02d}-{:02d}.d".format(self.cron_start.year,
                                                                          self.cron_start.month,
                                                                          self.cron_start.day))

    def _media_file(self, hms, day_dir=None):
        """ media_file - create new, unique media file based on time """

        _root = self.dest_dir
        if day_dir:
            _root = os.path.join(_root, day_dir)
            os.makedirs(_root, exist_ok=True)

        _cur_fname = os.path.join(_root, "{}-0000.{}".format(hms, self.suffix))
        self.logger.debug("NEW - new file {}".format(_cur_fname))
        if os.path.exists(_cur_fname):
            # rotating in under a second -- uniquify it
            self.ext += 1
            _cur_fname = os.path.join(_root, "{}-{:04}.{}".format(hms, self.ext, self.suffix))
            self.logger.debug("NEW - overlap iterate file {}".format(_cur_fname))
        else:
            self.ext = 0

        _cur_fd = self.open_media_file(_cur_fname)
        return _cur_fname, _cur_fd

    def _close_cur_file(self):
        """ close_cur_file -- UNLOCKED, close cur file and if file is full and compressing, signal compression """
        if self.media_file is not None:
            self.media_file = None

        if self.enable_compress and (self.cur_file_name is not None):
            # queue the previous output file for compression by the compression thread

            # making the management thread more robust
            if not self.file_work_thread.isAlive():
                self.logger.info("The management thread has died starting it again.")
                self.file_work_thread = ManagementThread(self, self.suffix)
                self.file_work_thread.start()

            with self.work_cond:
                if self.n_records >= self.max_records:
                    self.to_compress.append(self.cur_file_name)
                self.cur_file_name = None

                self.work_cond.notify()

    def _parse_to_ms(self, period_str):
        """ parse_to_ms - return seconds and ms for strings representing
                          weeks/days/hours/mins/secs/ms in periodic tasks """
        # general format is <number>[ ]unit[s][, ...]

        # pull out the spaces, lower case, make a list
        _period = period_str. \
            strip(). \
            lower(). \
            replace(' ', ''). \
            split(',')

        # 600M sec/week, track ms and seconds separately
        _total_sec = 0
        _total_ms = 0
        for p in _period:
            _m = re.match(r'^([0-9]+\.?[0-9]*)(ms|sec|s|min|m|hr|hour|day|dy|week|wk)s?$', p)
            if not _m:
                raise ValueError("{}: invalid period specification".format(p))
            _count, _unit = _m.groups()
            _count = float(_count) if '.' in _count else int(_count)
            if _unit in ['wk', 'week']:
                _total_sec += (7 * 24 * 60 * 60) * _count
            elif _unit in ['day', 'dy']:
                _total_sec += (24 * 60 * 60) * _count
            elif _unit in ['hour', 'hr']:
                _total_sec += (60 * 60) * _count
            elif _unit in ['min', 'm']:
                _total_sec += 60 * _count
            elif _unit in ['sec', 's']:
                # does seconds have a fractional portion?
                _frac = round(_count - int(_count), 6)
                if _frac != 0:
                    _total_ms += int(_frac * 1000000)
                    _count = int(_count)
                _total_sec += _count
            elif _unit == 'ms':
                _total_ms += _count

        return int(_total_sec), int(_total_ms)


class CSVHandler(MediaHandler):
    """
    CSV-specific logic for writing files
    """

    def __init__(self, config, suffix, ingest_ref, callback):
        super().__init__(config, suffix, ingest_ref, callback)

        self.fields = None

    def open_media_file(self, cur_file):
        _media_file = csv.writer(open(cur_file, 'w'))
        return _media_file

    def write_rows(self, obj):
        if self.media_file:
            self.media_file.writerows(obj)

    def extract_rows(self, data):
        # extract_rows - convert a list of dict's with readings into a series of ordered rows
        if data == []:
            return []

        # return a row of data in a standard field-order. assume you have a list of readings.

        for d in data:
            # if we're adding a timestamp field, must not be in the data -- take it from the body
            _ts = [d[self.ts_col]] if self.add_timestamp else []
            _readings = d['readings']
            if self.start_collection:
                yield [_readings[field] for field in self.fields] + _ts


class PickleHandler(MediaHandler):
    """ PickleHandler -- pkl-specific logic for writing pickle files """

    def __init__(self, config, suffix, ingest_ref, callback):
        super().__init__(config, suffix, ingest_ref, callback)

    def open_media_file(self, cur_file):
        _media_file = open(cur_file, 'wb')
        return _media_file

    def extract_rows(self, data):
        for d in data:
            yield d

    def write_rows(self, obj):
        pickle.dump(obj, self.media_file, pickle.HIGHEST_PROTOCOL)


#
# ------------------- PLUGIN API-----------------------
#
_DEFAULT_CONFIG = {
    "plugin": {  # mandatory  filter
        "description": "filter which writes its output to a series of csv files",
        "type": "string",
        "default": "csv_writer",
        "readonly": "true"
    },

    "inputAssets": {
        "description": "Comma separated list of input asset names (empty -> takes all)",
        "type": "string",
        "default": "",
        "displayName": "Input assets name",
        "order": "1"
    },
    "forwardData": {
        "description": "forward raw data independent of filter",
        "type": "boolean",
        "default": "false",
        "displayName": "Forward data",
        "order": "2"
    },
    # data storage options
    "destDir": {
        "description": "Destination directory inside $FOGLAMP_DATA "
                       "e.g. /usr/local/foglamp/data/readings-out if FOGLAMP_DATA/ is prefixed "
                       "otherwise it will be created as specified. Default is FOGLAMP_DATA/readings-out."
                       " If the path given without FOGLAMP_DATA/ then "
                       "directory  will be created inside $FOGLAMP_ROOT/services/<path>, path can be recursive.",
        "type": "string",
        "default": "{}readings-out".format(FOGLAMP_DATA_PFX),
        "displayName": "Destination directory",
        "order": "3"
    },
    "filterName": {
        "description": "Value under this field will create a sub-directory inside "
                       "the Destination Directory. E.g.,"
                       " Use <south service name>-storage if the filter is "
                       "applied between south service and storage.",
        "type": "string",
        "default": "south-storage",
        "displayName": "Subdirectory name",
        "order": "4"
    },
    "fileType": {
        "description": "Type of file in which readings to be stored.",
        "type": "enumeration",
        "default": "csv",
        "options": ["csv", "pickle"],
        "displayName": "File type",
        "order": "5"
    },
    "dynamicSamplingRate": {
        "description": "Whether data stream has dynamic (non-fixed) sampling rate",
        "type": "boolean",
        "default": "false",
        "displayName": "Dynamic sampling rate",
        "order": "6"
    },
    "samplingRate": {
        "description": "Incoming samples per second",
        "type": "integer",
        "default": "8000",
        "displayName": "Sampling rate",
        "order": "7",
        "validity": "dynamicSamplingRate == \"false\"",
    },
    "cronMode": {
        "description": "Cron style, either periodic, continuous, or table",
        "type": "enumeration",
        "options": ["continuous", "periodic", "table"],
        "default": "continuous",
        "displayName": "Cron mode",
        "order": "8"
    },
    # CRON CONTINUOUS and PERIODIC
    "cronPeriodStart": {
        "description": "Time (UTC, potentially in the past) cronPeriods are synchronized with. "
                       "In periodic mode, <cronPeriod> worth of records are written per file "
                       "every <cronPeriodCycle> units until cronTotalDuration is reached.. "
                       "In continuous mode, records are written continuously, filling files "
                       "of <cronPeriod> length until <cronTotalDuration> is reached. In this mode "
                       "record savings start immediately but the first file may comprise a partial period."
                       "Give a empty string to consider the start from the first reading. Otherwise give time stamp at "
                       "which the cron should start.  For example : 2021-04-27 09:25:35.300875+00:00 ",
        "type": "string",
        "default": "",
        "displayName": "Cron period starting time",
        "order": "9"
    },
    "cronPeriodDuration": {
        "description": "Amount of time records are written per file (in ms, sec[s], min[s], hr[s], day[s], week[s])",
        "type": "string",
        "default": "2mins",
        "displayName": "Recording period duration",
        "validity": "cronMode == \"continuous\"",
        "order": "10"
    },
    "eventRepetitionTime": {
        "description": "The time after which the desired event repeats [eg., 16mins]",
        "type": "string",
        "default": "16mins",
        "displayName": "Event repetition time",
        "validity": "cronMode == \"periodic\"",
        "order": "11"
    },
    "eventPreTime": {
        "description": "The amount of time to consider for collection"
                       " before the desired event [eg., 1min]",
        "type": "string",
        "default": "1min",
        "displayName": "Pre event time",
        "validity": "cronMode == \"periodic\"",
        "order": "12"
    },
    "eventDuration": {
        "description": "The duration for desired event [eg., 1min]",
        "type": "string",
        "default": "1min",
        "displayName": "Event duration",
        "validity": "cronMode == \"periodic\"",
        "order": "13"
    },
    "eventPostTime": {
        "description": "The amount of time to consider for collection"
                       " after the desired event [eg., 1min]",
        "type": "string",
        "default": "1min",
        "displayName": "Post event time",
        "validity": "cronMode == \"periodic\"",
        "order": "14"
    },
    "rotateAfter": {
        "description": "Total duration after which files are rotated (eg., 4wks",
        "type": "string",
        "default": "14days",
        "displayName": "Rotate after",
        "order": "15"
    },
    # CRON-TAB
    "cronTabSpec": {
        "description": "Modified cron format for when collection happens."
                       " \"*\" stands for first-last. Entry format: "
                       "minute(0-59) hour(0-23) day-of-month (1-31) month(1-12/names)"
                       " day-of-week(0-7 or names) duration(seconds[float]) "
                       "Comma-separated lists are allowed. Ranges are allowed. "
                       "If empty string, maxRecords/maxFiles are used.",
        "type": "string",
        "default": "",
        "displayName": "Periodic collection spec",
        "validity": "cronMode == \"table\"",
        "order": "16"
    },
    # CSV-specific options
    "addTimestamp": {
        "description": "Add timestamp to csv data",
        "type": "boolean",
        "default": "false",
        "displayName": "Add timestamp to csv data",
        "order": "17"
    },
    # compression-specific options
    "enableCompress": {
        "description": "Whether the csv files collected are to be compressed or not.",
        "type": "boolean",
        "default": "true",
        "displayName": "Enable compression",
        "validity": " fileType == \"csv\"",
        "order": "18"
    },
    "compressionType": {
        "description": "Compress output files",
        "type": "enumeration",
        "default": "bzip2",
        "options": ["bzip2", "gzip", "7za"],
        "displayName": "Compression type",
        "validity": " fileType == \"csv\"  && enableCompress == \"true\"",
        "order": "19"
    },
    "encryptPw": {
        "description": "Encrypt compressed file with this password",
        "type": "string",
        "default": "",
        "displayName": "Encryption password",
        "validity": " compressionType == \"7za\"",
        "order": "20"
    },
    "enable": {
        "description": "Enable csv writer plugin",
        "type": "boolean",
        "default": "true",
        "displayName": "Enabled",
        "order": "21"
    },
}


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.info("CSVW - plugin_info called")
    return {
        'name': 'CSVWriter',
        'version': '2.3.0',
        'mode': 'none',
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin.
    Args:
        config: JSON configuration document for the plugin configuration category
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """

    _LOGGER.info("PLUGIN INIT, making a new media handle")
    try:

        if config['fileType']['value'] == 'csv':
            _media_handle = CSVHandler(config, "csv", ingest_ref, callback)
        else:
            _media_handle = PickleHandler(config, "pkl", ingest_ref, callback)

        _handle_map = HandleMap(__name__)
        _h = _handle_map.new_handle(_media_handle, config)
        _LOGGER.info("type(_handle_map)={}, type(_h)={}".format(type(_handle_map), type(_h)))
        _h['handle_map'] = _handle_map
    except:
        _LOGGER.exception("Could not create media handle")
        raise

    return _h


def plugin_ingest(h, data):
    """ plugin_ingest -- log data we receive """

    if type(data) == dict:
        # make everything a list
        data = [data]

    _handle_map = h['handle_map']
    _media_handle = _handle_map.get_handle(h)
    _media_handle.ingest(data)


def plugin_reconfigure(h, new_config):
    """ Reconfigures the plugin
    Args:
        h: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _handle_map = h['handle_map']
    _handle = _handle_map.get_handle(h)
    _LOGGER.info("Old config for  plugin {} \n new config {}".format(h, new_config))
    ingest_ref, callback = _handle.ingest_ref, _handle.callback
    plugin_shutdown(h)
    # plugin_init
    new_handle = plugin_init(new_config, ingest_ref, callback)
    return new_handle


def plugin_shutdown(h):
    """ Shut down the plugin.
    Args:
        h: JSON configuration document for the plugin configuration category
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """
    _LOGGER.info("The csv writer plugin is going to shutdown.")
    _handle_map = h['handle_map']
    _media_handle = _handle_map.get_handle(h)
    _media_handle.shutdown()

    _handle_map.del_handle(h)  # remove our mapping
    _LOGGER.info("The csv writer plugin has successfully shutdown.")

